import Data.List
import Data.Maybe
import qualified Data.Map as Map
import Control.Parallel.Strategies(using, parList, rseq)
import System.Environment as System

data Othello = White | Black | Empty deriving (Eq, Show)
type P = (Int, Int)
type Board = Map.Map P Othello

--tested at depths of 2 and 4
--can adjust the depth used here manually:
depth_set = 4

colorToString :: Othello -> String
colorToString piece =
    case piece of
        Empty -> "_"
        White -> "w"
        Black -> "b"

setBoard :: Board -> String
setBoard b = "\n 0 1 2 3 4 5 6 7 \n" ++ (intercalate "\n" (map (setRow b)[0..7])) ++ "\n 0 1 2 3 4 5 6 7 \n \n"
    where
        setRow b r = show r ++ " " ++ (intercalate " " (map (\p -> 
            colorToString (fromMaybe Empty (Map.lookup p b)))
            ([ (r,x) | x <- [0..7]]))) ++ " " ++ show r

changeColor :: Othello -> Othello
changeColor White = Black
changeColor Black = White
changeColor _ = Empty

moves :: Othello -> Board -> [P]
moves col b = filter (isValid col b) possibilities
        where
            possibilities = [(x, y) | x <- [0..7], y <- [0..7]]

isValid :: Othello -> Board -> P -> Bool
isValid col b p = flipCell col b p /= [] && isNothing (Map.lookup p b)

move :: Othello -> Board -> P -> Board
move col b p = Map.union helper b
  where 
      helper = (Map.fromList (zip (flipCell col b p) (repeat col)))

adv :: Othello -> Board -> Int
adv col b = sum ( map (\(_, x) -> pt x)  (Map.toList b))
  where
    pt x
      | x == col = 1
      | x == Empty = 0
      | otherwise = -1

dir = [(0, 1), (1, 0), (1, 1), (0, -1), (-1, 0), (-1, -1), (-1, 1), (1, -1)]

flipCell :: Othello -> Board -> P -> [P]
flipCell col b p 
    | null flip = []
    | otherwise = p : flip
    where 
        flip = concat(map(flipCellHelper True col b p) dir)

flipCellHelper :: Bool -> Othello -> Board -> P -> P -> [P]
flipCellHelper i col b p d
    | nxtCol == changeColor col = if r /= [] then
        if not i then
            p : r
        else r 
    else []
    | nxtCol == col = if not i then [p]
    else []
    | otherwise = []
    where
        nxtP = ( \(x1,y1)(x2,y2) -> (x1+x2, y1+y2) ) p d
        nxtCol = fromMaybe Empty (Map.lookup nxtP b)
        r = flipCellHelper False col b nxtP d

endGame :: Othello -> Int -> IO ()
endGame col adv
    | adv == 0 = putStr "Tie!\n"
    | adv > 0 = putStr (colorToString col ++ " Won!\n")
    | otherwise = putStr (colorToString (changeColor col) ++ " Won!\n")


minimax :: Int -> Othello -> Board -> Int
minimax dpth col b
    | endGame = if (adv col b) > 0 
        then 100000
        else -100000
    | dpth <= 0 = adv col b
    | otherwise = if ( moves (changeColor col) b ) /= []
        then -maxPt
        else maxPt
    where
        endGame = null (moves col b) && null (moves (changeColor col) b)
        clrUp = if ( moves (changeColor col) b ) /= []
            then changeColor col
            else col
        nm = if clrUp /= col
            then ( moves (changeColor col) b )
            else ( moves col b)
        maxPt = maximum (map (minimax (dpth - 1) clrUp . move clrUp b) nm `using` parList rseq)

--computerMove is where depth is set: default is 4
play :: Othello -> Board -> IO ()
play col b = 
    if null (moves col b) && null (moves (changeColor col) b)
        then endGame col $ adv col b 
        else do
            let
                brd = move col b (computerMove col b)
                clrUp = if moves (changeColor col) brd /= []
                    then (changeColor col)
                    else col
            putStr (setBoard b)
            play clrUp brd 
            where computerMove col b = fst $ maximumBy (\(_, x) (_, y) -> compare x y)(map (\p -> (p, minimax depth_set col (move col b p)))(moves col b))

--board1 is the default start state: 4 pieces in the center, 2 white, 2 black
board1 = Map.fromList [((3, 3), White), ((4, 4), White), ((3, 4), Black), ((4, 3), Black)]

--board2 is a potential mid-game scenario which is black-piece dominated
--having 12 black pieces and 6 white pieces
board2 = Map.fromList [((2,2), Black), ((2,3), Black), ((2,4), Black), ((3,2), Black), ((3,3), Black), ((3,4), Black), ((4,1), Black), ((4,3), Black), ((4,4), Black), ((4,6), Black), ((6,3), Black), ((3,1), White), ((4,2), White), ((4,5), White), ((5,3), White), ((5,6), White),((6,4), White) ]

--board3 is a potential mid-game scenario, largely white-piece dominant
--totaling 11 white pieces, and 7 black pieces
board3 = Map.fromList [((1,2), Black), ((2,2), Black), ((2,3), Black), ((3,4), Black), ((4,2), Black), ((4,5), Black), ((5,5), Black), ((1,3), White), ((2,4), White), ((2,5), White), ((2,6), White), ((3,2), White), ((3,3), White), ((3,5), White), ((3,6), White), ((4,3), White), ((4,4), White), ((5,3), White)]

main :: IO ()
main = do
    args <- System.getArgs
    if length args == 0
        then play White board1
        else do 
            if head args == "board1"
                then play White board1
                else do
                    if head args == "board2"
                        then play White board2
                        else do
                            if head args == "board3"
                                then play White board3
                                else do putStr "invalid usage"
